                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _PROCEDURALTEXTURE_H_
#define _PROCEDURALTEXTURE_H_

#include "../3D/Vertex.h"
#include "../3D/PerlinNoise.h"

typedef void (* PROCEDURALTEXTUREFUNCTION) (unsigned char *&pixels, unsigned int &width, unsigned int &height, __int64 timeStamp);

/* ################################################################################################################### */


class ProceduralTexture : public Texture {
protected:
	unsigned char *pixels;
    PROCEDURALTEXTUREFUNCTION proceduralFunc;
    __int64 currentTime, lastTime;
public:
	ProceduralTexture(PROCEDURALTEXTUREFUNCTION ProceduralFunc);
	virtual ~ProceduralTexture();

	bool upload();
    bool update(__int64 timeStamp);
    bool activate(int textureUnit);
};

/* ----------------------------------------------------*/

ProceduralTexture::ProceduralTexture(PROCEDURALTEXTUREFUNCTION ProceduralFunc){
	pixels = NULL;
    width = height = 0;
    depth  = 1;
    channels = 4;
    proceduralFunc = ProceduralFunc;
    currentTime = lastTime = 0;

	textureID = NULL;
	minFilter = D3DTEXF_LINEAR;
	magFilter = D3DTEXF_LINEAR;
	mipFilter = D3DTEXF_POINT;
	wrapModeS = D3DTADDRESS_WRAP;
	wrapModeT = D3DTADDRESS_WRAP;
	wrapModeR = D3DTADDRESS_WRAP;
	internalFormat = D3DFMT_A8R8G8B8;
}

ProceduralTexture::~ProceduralTexture(){
}


bool ProceduralTexture::upload(){
	//d3ddevice->CreateTexture(width, height, 1, D3DUSAGE_DYNAMIC, internalFormat, D3DPOOL_DEFAULT, &textureID);

    proceduralFunc(pixels, width, height, currentTime);
	d3ddevice->CreateTexture(width, height, 1, 0, internalFormat, D3DPOOL_MANAGED, (LPDIRECT3DTEXTURE8 *) &textureID);

	
	return true;
}

bool ProceduralTexture::update(__int64 timeStamp){
    currentTime = timeStamp;
    return true;
}

bool ProceduralTexture::activate(int textureUnit){
    if (Texture::activate(textureUnit)){
        if (currentTime != lastTime){
            proceduralFunc(pixels, width, height, currentTime);
            if (pixels != NULL){
				
				D3DLOCKED_RECT rect;
				if (((LPDIRECT3DTEXTURE8) textureID)->LockRect(0, &rect, NULL, 0) == D3D_OK){
					memcpy(rect.pBits, pixels, width * height * channels);
					((LPDIRECT3DTEXTURE8) textureID)->UnlockRect(0);
				}

			}
            currentTime = lastTime;
        }
        return true;
    }
    return false;
}


void MakeFlameTexture(unsigned char *&pixels, unsigned int &width, unsigned int &height, __int64 timeStamp){
    static unsigned int flamePal[256] = {
        0x000000, 0x100000, 0x140000, 0x180000, 0x1C0000, 0x200000, 0x240000, 0x280000, 0x280008, 0x240010, 0x240018, 0x200020, 0x1C0028, 0x1C0030, 0x180038, 0x140040,
        0x140048, 0x100050, 0x100058, 0x0C0060, 0x080068, 0x080070, 0x040078, 0x000080, 0x000082, 0x000084, 0x000088, 0x00008C, 0x000090, 0x000092, 0x000094, 0x000098,
        0x00009C, 0x0000A0, 0x0000A0, 0x0000A4, 0x0000A8, 0x0000AC, 0x0000B0, 0x0000B4, 0x0004B8, 0x0004BC, 0x0008C0, 0x0008C4, 0x000CC8, 0x000CCC, 0x0010D0, 0x0010D4,
        0x0014D8, 0x0014DC, 0x0018E0, 0x0018E4, 0x001CE8, 0x001CEC, 0x0020F0, 0x0020F4, 0x0024FC, 0x0026FC, 0x0028FC, 0x002AFC, 0x002CFC, 0x002EFC, 0x0030FC, 0x0032FC,
        0x0034FC, 0x0036FC, 0x0038FC, 0x003AFC, 0x003CFC, 0x003EFC, 0x0040FC, 0x0042FC, 0x0044FC, 0x0046FC, 0x0048FC, 0x004AFC, 0x004CFC, 0x004EFC, 0x0050FC, 0x0052FC,
        0x0054FC, 0x0056FC, 0x0058FC, 0x005AFC, 0x005CFC, 0x0060FC, 0x0062FC, 0x0064FC, 0x0066FC, 0x0068FC, 0x006AFC, 0x006CFC, 0x006EFC, 0x0070FC, 0x0072FC, 0x0074FC,
        0x0076FC, 0x0078FC, 0x007AFC, 0x007CFC, 0x007EFC, 0x0080FC, 0x0082FC, 0x0084FC, 0x0086FC, 0x0088FC, 0x008AFC, 0x008CFC, 0x008EFC, 0x0090FC, 0x0092FC, 0x0095FC,
        0x0098FC, 0x009AFC, 0x009CFC, 0x009EFC, 0x00A0FC, 0x00A2FC, 0x00A4FC, 0x00A6FC, 0x00A8FC, 0x00AAFC, 0x00ACFC, 0x00AEFC, 0x00B0FC, 0x00B2FC, 0x00B4FC, 0x00B6FC,
        0x00B8FC, 0x00BAFC, 0x00BCFC, 0x00BEFC, 0x00C0FC, 0x00C2FC, 0x00C4FC, 0x00C6FC, 0x00C8FC, 0x00CAFC, 0x00CCFC, 0x00CEFC, 0x00D0FC, 0x00D1FC, 0x00D2FC, 0x00D3FC,
        0x00D4FC, 0x00D5FC, 0x00D6FC, 0x00D7FC, 0x00D8FC, 0x00D9FC, 0x00DAFC, 0x00DBFC, 0x00DCFC, 0x00DDFC, 0x00DEFC, 0x00DFFC, 0x00DFFC, 0x00E0FC, 0x00E1FC, 0x00E2FC,
        0x00E3FC, 0x00E4FC, 0x00E5FC, 0x00E6FC, 0x00E7FC, 0x00E8FC, 0x00E9FC, 0x00EAFC, 0x00EBFC, 0x00ECFC, 0x00EDFC, 0x00EEFC, 0x00EFFC, 0x00EFFD, 0x00F0FC, 0x00F1FC,
        0x00F2FC, 0x00F3FC, 0x00F4FC, 0x00F5FC, 0x00F6FC, 0x00F7FC, 0x00F8FC, 0x00F9FC, 0x04FAFC, 0x80FBFC, 0x0CFCFC, 0x10FDFC, 0x14FCFC, 0x18FCFC, 0x1CFCFC, 0x20FCFC,
        0x24FCFC, 0x28FCFC, 0x2AFCFC, 0x2CFCFC, 0x30FCFC, 0x34FCFC, 0x38FCFC, 0x3CFCFC, 0x40FCFC, 0x44FCFC, 0x48FCFC, 0x4CFCFC, 0x50FCFC, 0x54FCFC, 0x56FCFC, 0x58FCFC,
        0x5CFCFC, 0x60FCFC, 0x64FCFC, 0x68FCFC, 0x6CFCFC, 0x70FCFC, 0x74FCFC, 0x78FCFC, 0x7CFCFC, 0x7EFCFC, 0x80FCFC, 0x84FCFC, 0x88FCFC, 0x8CFCFC, 0x90FCFC, 0x94FCFC,
        0x98FCFC, 0x9EFCFC, 0xA0FCFC, 0xA4FCFC, 0xA8FCFC, 0xAAFCFC, 0xACFCFC, 0xB0FCFC, 0xB4FCFC, 0xB8FCFC, 0xBCFCFC, 0xC0FCFC, 0xC4FCFC, 0xC8FCFC, 0xCCFCFC, 0xD0FCFC,
        0xD2FCFC, 0xD4FCFC, 0xD8FCFC, 0xDCFCFC, 0xE0FCFC, 0xE4FCFC, 0xE8FCFC, 0xECFCFC, 0xF0FCFC, 0xF4FCFC, 0xF8FCFC, 0xFCFCFC, 0xFDFCFC, 0xFEFCFC, 0xFFFCFC, 0xFFFFFF
    };

    static unsigned char flame[129 * 128];
    static unsigned int fire[128 * 128];

    static char sinus[128];
    static char rands[128];
	static __int64 lTimeStamp = 0;
	static int t = 0;

    width = height = 128;
    if (timeStamp - lTimeStamp < 16000){
        pixels = NULL;
        return;
    }

	lTimeStamp = timeStamp;

    if (t == 0){
        for (int i = 0; i < 128; i++){
		    sinus[i] = (int) (1.5 * cosf(i * PI / 64));
            rands[i] = rand() & 0xFF;
        }
    }
    pixels = (unsigned char *) fire;


/*    unsigned int x,y,a;
    for (x = 16384 + 32; x < 16512-32; x++){
        a = rand() & 0xFF;
        flame[x] = (a < 96)? flame[x-1] : (a & 1) - 1;
    }

    for (y = 0; y < 128; y++){
        for (x = 0; x < 128; x++){
            flame[y * 128 + x] = 
                (flame[(y + 0) * 128 + x] + 
                 flame[(y + 0) * 128 + ((x - 1) & 127)] + 
                 flame[(y + 0) * 128 + ((x + 1) & 127)] + 
                 flame[(y + 1) * 128 + x]) >> 2;
        }
    }

    for (y = 0; y < 128; y++){
        for (x = 0; x < 128; x++){
            fire[y * 128 + x] = flamePal[flame[y * 128 + x]];
        }
    }

    return;
    */

	__asm {
		LEA		EDI, [flame]
		XOR		ESI, ESI
		XOR		EBX, EBX
        MOV     ECX, t
        MakeBottomLine:
            IMUL    ECX, 422423
            MOV     EAX, ECX
            SHR     EAX, 24
            AND     EAX, 127
            MOV     AL, [rands + EAX]

			CMP		AL, 96
            JA		DrawNext
				TEST	AL, 1
				SETZ	BL
				DEC		BL
			DrawNext:

			MOV		[EDI + ESI + 16384], BL
			INC		ESI
			CMP		ESI, 128
        JNZ MakeBottomLine

		MOV		ECX, 1
		XOR		EDX, EDX
		ADD		t, 48
				
		DrawFlameY:
			DrawFlameX:
				MOV		EAX, EDX
				MOV		EBX, EDX
				MOV		ESI, ECX
				INC		EAX
				DEC		EBX
				SHL		ESI, 7
				AND		EAX, 127
				AND		EBX, 127
				ADD		EAX, ESI
				ADD		EBX, ESI
				MOVZX	EAX, BYTE PTR [EDI + EAX]
				MOVZX	EBX, BYTE PTR [EDI + EBX]
				ADD		ESI, EDX
				ADD		EAX, EBX
				MOVZX	EBX, BYTE PTR [EDI + ESI]
				INC		ESI
				ADD		EAX, EBX
				MOVZX	EBX, BYTE PTR [EDI + ESI + 127]
				SUB		ESI, EDX
				ADD		EAX, EBX
				ADD		ESI, -128
				SHR		EAX, 2
				MOV		EBX, ECX
				SETNZ	AH
				IMUL	EBX, EBX
				SUB		AL, AH

				ADD		EBX, t
				DEC		ESI
				SHR		EBX, 5
				AND		EBX, 127
				MOVZX	EAX, AL
				MOVZX	EBX, BYTE PTR [sinus + EBX]
				ADD		EBX, EDX
				AND		EBX, 127
				INC		EDX
				ADD		ESI, EBX
			
				CMP		EDX, 127
				MOV		[EDI + ESI], AL
				LEA		EBX, [fire + 4 * ESI]
				MOV		EAX, [flamePal + 4 * EAX]

				MOV		[EBX], EAX

			JNA DrawFlameX
			INC		ECX
			XOR		EDX, EDX
			CMP		ECX, 128
		JNA		DrawFlameY
	}
}


void MakePerlinNoiseTexture(unsigned char *&pixels, unsigned int &width, unsigned int &height, __int64 timeStamp){
    static unsigned char noise[128 * 128 * 4];

	static __int64 lTimeStamp = 0;

    width = height = 128;
    if (timeStamp - lTimeStamp < 16000){
        pixels = NULL;
        return;
    }
	if (lTimeStamp == 0){
		initPerlin(0);		
	}

	lTimeStamp = timeStamp;

	float x,y,z;
	Vertex v;

	unsigned char *dest = noise;
	for (int i = 0; i < 128; i++){
		for (int j = 0; j < 128; j++){
			x = (i - 63.5f) * 0.132f;
			y = (j - 63.5f) * 0.132f;
			z = timeStamp * 0.000001f;
			v.x = noise3(x,y,z);

			x = (i - 63.5f) * 0.212f;
			y = (j - 63.5f) * 0.212f;
			z = timeStamp * 0.000001f;
			v.y = noise3(x,y,z);

			x = (i - 63.5f) * 0.212f;
			y = (j - 63.5f) * 0.212f;
			z = timeStamp * 0.000001f;
			v.z = noise3(x,y,z);
			//v.z = 1;
			
			v.fastNormalize();

			*dest++ = unsigned char(127.5f * (v.x + 1));
			*dest++ = unsigned char(127.5f * (v.y + 1));
			*dest++ = unsigned char(127.5f * (v.z + 1));
			*dest++ = 128;
		
		}
	}


    pixels = (unsigned char *) noise;
}

#endif // _PROCEDURALTEXTURE_H_