                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _CUBEMAP_H_
#define _CUBEMAP_H_

/* ################################################################################################################### */


class CubeMap : public Texture {
protected:
	unsigned char *pixels[6];
	unsigned int format;
public:
	CubeMap();
	virtual ~CubeMap();

	bool loadFromFiles(const char *filePosX, const char *fileNegX, const char *filePosY, const char *fileNegY, const char *filePosZ, const char *fileNegZ);
    void loadNormalizing(unsigned int size);

	bool upload();
};

/* ----------------------------------------------------*/

CubeMap::CubeMap(){
	for (int i = 0; i < 6; i++) pixels[i] = NULL;
	width = height = depth = channels = format = 0;
	textureID = TEXTURE_INVALID;
	minFilter = GL_LINEAR_MIPMAP_LINEAR;
	magFilter = GL_LINEAR;
	wrapModeS = GL_CLAMP_TO_EDGE_EXT;
	wrapModeT = GL_CLAMP_TO_EDGE_EXT;
	wrapModeR = GL_CLAMP_TO_EDGE_EXT;
	internalFormat = GL_RGB8;
	target = GL_TEXTURE_CUBE_MAP_ARB;
}

CubeMap::~CubeMap(){
    for (int i = 0; i < 6; i++){
	    if (pixels[i] != NULL) delete pixels[i];
    }
}

bool CubeMap::loadFromFiles(const char *filePosX, const char *fileNegX, const char *filePosY, const char *fileNegY, const char *filePosZ, const char *fileNegZ){
    bool ok = true;

    for (int i = 0; i < 6; i++){
    	if (pixels[i] != NULL){
	    	delete pixels[i];
		    pixels[i] = NULL;
    	}
    }
    ok &= (loadImage(filePosX, pixels[0], width, height, channels, format) == IMAGE_SUCCESSFUL);
    ok &= (loadImage(fileNegX, pixels[1], width, height, channels, format) == IMAGE_SUCCESSFUL);
    ok &= (loadImage(filePosY, pixels[2], width, height, channels, format) == IMAGE_SUCCESSFUL);
    ok &= (loadImage(fileNegY, pixels[3], width, height, channels, format) == IMAGE_SUCCESSFUL);
    ok &= (loadImage(filePosZ, pixels[4], width, height, channels, format) == IMAGE_SUCCESSFUL);
    ok &= (loadImage(fileNegZ, pixels[5], width, height, channels, format) == IMAGE_SUCCESSFUL);

    if (ok) depth = 1;

    return ok;
}

unsigned char PackFloatInByte(float x){
    return (unsigned char) (127.5f * (x + 1));
}

void CubeMap::loadNormalizing(unsigned int size){
    height = width = size;
    depth = 1;
    channels = 3;
    format = IMAGE_FORMAT_UNCOMPRESSED;

	unsigned int i,x,y,z;
	GLfloat fX, fY, fZ, oolen;
	unsigned char *pixel;

	channels = 3;
	float halfsize = size / 2 - 0.5f;

	for (i = 0; i < 6; i++){
    	if (pixels[i] != NULL){
	    	delete pixels[i];
		    pixels[i] = NULL;
    	}
		pixels[i] = (unsigned char *) malloc(3 * size * size);
	}
	
	pixel = pixels[0];
	// Positive X face
	for (y = 0; y < size; y++) {
		for (z = 0; z < size; z++) {
			fX = (GLfloat) (halfsize);
			fY = (GLfloat) (halfsize - y);
			fZ = (GLfloat) (halfsize - z);
			oolen = 1.0f/sqrtf(fX*fX + fY*fY + fZ*fZ);
			fX *= oolen;  fY *= oolen;  fZ *= oolen;
			pixel[y*3*size + z*3 + 0] = PackFloatInByte(fX);
			pixel[y*3*size + z*3 + 1] = PackFloatInByte(fY);
			pixel[y*3*size + z*3 + 2] = PackFloatInByte(fZ);
		}
	}
	
	pixel = pixels[1];
	// Negative X face
	for (y = 0; y < size; y++) {
		for (z = 0; z < size; z++) {
			fX = (GLfloat) (-halfsize);
			fY = (GLfloat) (halfsize - y);
			fZ = (GLfloat) (z - halfsize);
			oolen = 1.0f/sqrtf(fX*fX + fY*fY + fZ*fZ);
			fX *= oolen;  fY *= oolen;  fZ *= oolen;
			pixel[y*3*size + z*3 + 0] = PackFloatInByte(fX);
			pixel[y*3*size + z*3 + 1] = PackFloatInByte(fY);
			pixel[y*3*size + z*3 + 2] = PackFloatInByte(fZ);
		}
	}
	
	pixel = pixels[2];
	// Positive Y face
	for (z = 0; z < size; z++) {
		for (x = 0; x < size; x++) {
			fX = (GLfloat) (x - halfsize);
			fY = (GLfloat) (halfsize);
			fZ = (GLfloat) (z - halfsize);
			oolen = 1.0f/sqrtf(fX*fX + fY*fY + fZ*fZ);
			fX *= oolen;  fY *= oolen;  fZ *= oolen;
			pixel[z*3*size + x*3 + 0] = PackFloatInByte(fX);
			pixel[z*3*size + x*3 + 1] = PackFloatInByte(fY);
			pixel[z*3*size + x*3 + 2] = PackFloatInByte(fZ);
		}
	}
	
	pixel = pixels[3];
	// Negative Y face
	for (z = 0; z < size; z++) {
		for (x = 0; x < size; x++) {
			fX = (GLfloat) (x - halfsize);
			fY = (GLfloat) (-halfsize);
			fZ = (GLfloat) (halfsize - z);
			oolen = 1.0f/sqrtf(fX*fX + fY*fY + fZ*fZ);
			fX *= oolen;  fY *= oolen;  fZ *= oolen;
			pixel[z*3*size + x*3 + 0] = PackFloatInByte(fX);
			pixel[z*3*size + x*3 + 1] = PackFloatInByte(fY);
			pixel[z*3*size + x*3 + 2] = PackFloatInByte(fZ);
		}
	}
	
	pixel = pixels[4];
	// Positive Z face
	for (y = 0; y < size; y++) {
		for (x = 0; x < size; x++) {
			fX = (GLfloat) (x - halfsize);
			fY = (GLfloat) (halfsize - y);
			fZ = (GLfloat) (halfsize);
			oolen = 1.0f/sqrtf(fX*fX + fY*fY + fZ*fZ);
			fX *= oolen;  fY *= oolen;  fZ *= oolen;
			pixel[y*3*size + x*3 + 0] = PackFloatInByte(fX);
			pixel[y*3*size + x*3 + 1] = PackFloatInByte(fY);
			pixel[y*3*size + x*3 + 2] = PackFloatInByte(fZ);
		}
	}
	
	pixel = pixels[5];
	// Negative Z face
	for (y = 0; y < size; y++) {
		for (x = 0; x < size; x++) {
			fX = (GLfloat) (halfsize - x);
			fY = (GLfloat) (halfsize - y);
			fZ = (GLfloat) (-halfsize);
			oolen = 1.0f/sqrtf(fX*fX + fY*fY + fZ*fZ);
			fX *= oolen;  fY *= oolen;  fZ *= oolen;
			pixel[y*3*size + x*3 + 0] = PackFloatInByte(fX);
			pixel[y*3*size + x*3 + 1] = PackFloatInByte(fY);
			pixel[y*3*size + x*3 + 2] = PackFloatInByte(fZ);
		}
	}
	

}


bool CubeMap::upload(){
    bool ok = true;

	if (textureID == TEXTURE_INVALID){
		glGenTextures(1, &textureID);
	}
	glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, textureID);

	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, minFilter);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, magFilter);

	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, wrapModeS);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, wrapModeT);
//	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_R_EXT, wrapModeR);

	//glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAX_ANISOTROPY_EXT, 1);
	//glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);

    for (int i = 0; i < 6; i++){
        //
        if (pixels[i] == NULL){
            ok = false;
            continue;
        }

        upload2DImage(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB + i, pixels[i], width, height, channels, format, internalFormat, minFilter);

	    delete pixels[i];
	    pixels[i] = NULL;
    }
	
	return ok;
}


#endif // _CUBEMAP_H_