                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


/*
    Much of the content of this file is based on work by Mark Morley
    http://www.markmorley.com/opengl/frustumculling.html
*/

#ifndef _FRUSTUM_H_
#define _FRUSTUM_H_

#define FRUSTUM_LEFT   0
#define FRUSTUM_RIGHT  1
#define FRUSTUM_BOTTOM 2
#define FRUSTUM_TOP    3
#define FRUSTUM_FAR    4
#define FRUSTUM_NEAR   5

class Frustum {
protected:
    float frustum[6][4];
    Vertex points[8];

    void createPoint(int index, int plane0, int plane1, int plane2);

public:
    bool pointInFrustum(const Vertex &v);
    bool sphereInFrustum(const Vertex &v, float radius);
    bool cubeInFrustum(const Vertex &v, float size);

    bool polygonInFrustum(const Vertex *vertices, const int nVertices);

    int planeIntersectsFrustum(Vertex normal, float offset);

    void updateFrustum();

/*    void draw(){
        glColor3f(1,1,1);
        glBegin(GL_LINE_LOOP);
        for (int i = 4; i < 8; i++){
            glVertex3fv((float *) (points + i));
        }
        glEnd();
    }
*/
};

/* ----------------------------------- */

void Frustum::createPoint(int index, int plane0, int plane1, int plane2){
    Vertex g0(frustum[plane0][0], frustum[plane1][0], frustum[plane2][0]);
    Vertex g1(frustum[plane0][1], frustum[plane1][1], frustum[plane2][1]);
    Vertex g2(frustum[plane0][2], frustum[plane1][2], frustum[plane2][2]);
    Vertex p (frustum[plane0][3], frustum[plane1][3], frustum[plane2][3]);

    float x,y,z, M[9];
    float invDetM = 1.0f / (g0.x * (g1.y * g2.z - g1.z * g2.y) -
                            g1.x * (g0.y * g2.z - g0.z * g2.y) + 
                            g2.x * (g0.y * g1.z - g0.z * g1.y));

    M[0] =  (g1.y * g2.z - g1.z * g2.y) * invDetM;
    M[1] = -(g1.x * g2.z - g1.z * g2.x) * invDetM;
    M[2] =  (g1.x * g2.y - g1.y * g2.x) * invDetM;

    M[3] = -(g0.y * g2.z - g0.z * g2.y) * invDetM;
    M[4] =  (g0.x * g2.z - g0.z * g2.x) * invDetM;
    M[5] = -(g0.x * g2.y - g0.y * g2.x) * invDetM;

    M[6] =  (g0.y * g1.z - g0.z * g1.y) * invDetM;
    M[7] = -(g0.x * g1.z - g0.z * g1.x) * invDetM;
    M[8] =  (g0.x * g1.y - g0.y * g1.x) * invDetM;

    x = M[0] * p.x + M[1] * p.y + M[2] * p.z;
    y = M[3] * p.x + M[4] * p.y + M[5] * p.z;
    z = M[6] * p.x + M[7] * p.y + M[8] * p.z;

    points[index] = -Vertex(x,y,z);
}

/* ----------------------------------- */

bool Frustum::pointInFrustum(const Vertex &v){
    for (int i = 0; i < 6; i++){
        if (frustum[i][0] * v.x + frustum[i][1] * v.y + frustum[i][2] * v.z + frustum[i][3] <= 0) return false;
    }
    return true;
}

bool Frustum::sphereInFrustum(const Vertex &v, float radius){
    for (int i = 0; i < 6; i++){
        if (frustum[i][0] * v.x + frustum[i][1] * v.y + frustum[i][2] * v.z + frustum[i][3] <= -radius) return false;
    }
    return true;
}

bool Frustum::cubeInFrustum(const Vertex &v, float size){
    for (int i = 0; i < 6; i++){
        if (frustum[i][0] * (v.x - size) + frustum[i][1] * (v.y - size) + frustum[i][2] * (v.z - size) + frustum[i][3] > 0) continue;
        if (frustum[i][0] * (v.x + size) + frustum[i][1] * (v.y - size) + frustum[i][2] * (v.z - size) + frustum[i][3] > 0) continue;
        if (frustum[i][0] * (v.x - size) + frustum[i][1] * (v.y + size) + frustum[i][2] * (v.z - size) + frustum[i][3] > 0) continue;
        if (frustum[i][0] * (v.x + size) + frustum[i][1] * (v.y + size) + frustum[i][2] * (v.z - size) + frustum[i][3] > 0) continue;
        if (frustum[i][0] * (v.x - size) + frustum[i][1] * (v.y - size) + frustum[i][2] * (v.z + size) + frustum[i][3] > 0) continue;
        if (frustum[i][0] * (v.x + size) + frustum[i][1] * (v.y - size) + frustum[i][2] * (v.z + size) + frustum[i][3] > 0) continue;
        if (frustum[i][0] * (v.x - size) + frustum[i][1] * (v.y + size) + frustum[i][2] * (v.z + size) + frustum[i][3] > 0) continue;
        if (frustum[i][0] * (v.x + size) + frustum[i][1] * (v.y + size) + frustum[i][2] * (v.z + size) + frustum[i][3] > 0) continue;
        return false;
    }
    return true;
}

bool Frustum::polygonInFrustum(const Vertex *vertices, const int nVertices){
    int i,j;

    for (i = 0; i < 6; i++){
        for (j = 0; j < nVertices; i++){
            if (frustum[i][0] * vertices[j].x + frustum[i][1] * vertices[j].y + frustum[i][2] * vertices[j].z + frustum[i][3] > 0) break;
        }
        if (j == nVertices) return false;
    }

    return true;
}

int Frustum::planeIntersectsFrustum(Vertex normal, float offset){
    int i, neg, pos;

    neg = pos = 0;
    for (i = 0; i < 8; i++){
        if (points[i] * normal + offset < 0) neg++; else pos++;
    }
    if (pos){
        if (neg) return 0; else return 1;
    } else return -1;
}

void Frustum::updateFrustum(){
    float proj[16];
    float modl[16];
    float clip[16];
    float t;

    /* Get the current PROJECTION matrix from OpenGL */
    glGetFloatv(GL_PROJECTION_MATRIX, proj);

    /* Get the current MODELVIEW matrix from OpenGL */
    glGetFloatv(GL_MODELVIEW_MATRIX, modl);

    /* Combine the two matrices (multiply projection by modelview) */
    clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
    clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
    clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
    clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

    clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
    clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
    clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
    clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

    clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
    clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
    clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
    clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];

    clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
    clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
    clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
    clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];

    /* Extract the numbers for the RIGHT plane */
    frustum[0][0] = clip[ 3] - clip[ 0];
    frustum[0][1] = clip[ 7] - clip[ 4];
    frustum[0][2] = clip[11] - clip[ 8];
    frustum[0][3] = clip[15] - clip[12];

    /* Normalize the result */
    t = 1.0f / sqrtf(frustum[0][0] * frustum[0][0] + frustum[0][1] * frustum[0][1] + frustum[0][2] * frustum[0][2]);
    frustum[0][0] *= t;
    frustum[0][1] *= t;
    frustum[0][2] *= t;
    frustum[0][3] *= t;

    /* Extract the numbers for the LEFT plane */
    frustum[1][0] = clip[ 3] + clip[ 0];
    frustum[1][1] = clip[ 7] + clip[ 4];
    frustum[1][2] = clip[11] + clip[ 8];
    frustum[1][3] = clip[15] + clip[12];

    /* Normalize the result */
    t = 1.0f / sqrtf(frustum[1][0] * frustum[1][0] + frustum[1][1] * frustum[1][1] + frustum[1][2] * frustum[1][2]);
    frustum[1][0] *= t;
    frustum[1][1] *= t;
    frustum[1][2] *= t;
    frustum[1][3] *= t;

    /* Extract the BOTTOM plane */
    frustum[2][0] = clip[ 3] + clip[ 1];
    frustum[2][1] = clip[ 7] + clip[ 5];
    frustum[2][2] = clip[11] + clip[ 9];
    frustum[2][3] = clip[15] + clip[13];

    /* Normalize the result */
    t = 1.0f / sqrtf(frustum[2][0] * frustum[2][0] + frustum[2][1] * frustum[2][1] + frustum[2][2] * frustum[2][2]);
    frustum[2][0] *= t;
    frustum[2][1] *= t;
    frustum[2][2] *= t;
    frustum[2][3] *= t;

    /* Extract the TOP plane */
    frustum[3][0] = clip[ 3] - clip[ 1];
    frustum[3][1] = clip[ 7] - clip[ 5];
    frustum[3][2] = clip[11] - clip[ 9];
    frustum[3][3] = clip[15] - clip[13];

    /* Normalize the result */
    t = 1.0f / sqrtf(frustum[3][0] * frustum[3][0] + frustum[3][1] * frustum[3][1] + frustum[3][2] * frustum[3][2]);
    frustum[3][0] *= t;
    frustum[3][1] *= t;
    frustum[3][2] *= t;
    frustum[3][3] *= t;

    /* Extract the FAR plane */
    frustum[4][0] = clip[ 3] - clip[ 2];
    frustum[4][1] = clip[ 7] - clip[ 6];
    frustum[4][2] = clip[11] - clip[10];
    frustum[4][3] = clip[15] - clip[14];

    /* Normalize the result */
    t = 1.0f / sqrtf(frustum[4][0] * frustum[4][0] + frustum[4][1] * frustum[4][1] + frustum[4][2] * frustum[4][2]);
    frustum[4][0] *= t;
    frustum[4][1] *= t;
    frustum[4][2] *= t;
    frustum[4][3] *= t;

    /* Extract the NEAR plane */
    frustum[5][0] = clip[ 3] + clip[ 2];
    frustum[5][1] = clip[ 7] + clip[ 6];
    frustum[5][2] = clip[11] + clip[10];
    frustum[5][3] = clip[15] + clip[14];

    /* Normalize the result */
    t = 1.0f / sqrtf(frustum[5][0] * frustum[5][0] + frustum[5][1] * frustum[5][1] + frustum[5][2] * frustum[5][2]);
    frustum[5][0] *= t;
    frustum[5][1] *= t;
    frustum[5][2] *= t;
    frustum[5][3] *= t;


    createPoint(0, FRUSTUM_NEAR, FRUSTUM_TOP,    FRUSTUM_LEFT);
    createPoint(1, FRUSTUM_NEAR, FRUSTUM_TOP,    FRUSTUM_RIGHT);
    createPoint(2, FRUSTUM_NEAR, FRUSTUM_BOTTOM, FRUSTUM_RIGHT);
    createPoint(3, FRUSTUM_NEAR, FRUSTUM_BOTTOM, FRUSTUM_LEFT);
    createPoint(4, FRUSTUM_FAR,  FRUSTUM_TOP,    FRUSTUM_LEFT);
    createPoint(5, FRUSTUM_FAR,  FRUSTUM_TOP,    FRUSTUM_RIGHT);
    createPoint(6, FRUSTUM_FAR,  FRUSTUM_BOTTOM, FRUSTUM_RIGHT);
    createPoint(7, FRUSTUM_FAR,  FRUSTUM_BOTTOM, FRUSTUM_LEFT);
}

#endif // _FRUSTUM_H_