                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _RENDERTEXTURE_H_
#define _RENDERTEXTURE_H_

class RenderTexture : public Texture {
protected:
	HDC hdc;
	HGLRC hglrc;

	HPBUFFERARB hPBuffer;
	HDC hPdc;
	HGLRC hPrc;

	bool create();
public:
	RenderTexture(int Width, int Height, bool wantCubeMap);
	virtual ~RenderTexture();

	virtual bool upload();
	virtual bool unload();

	void beginRenderToTexture();
	void endRenderToTexture();
	void setRenderFace(GLint face);

	virtual bool activate();
	virtual bool deactivate();
};


RenderTexture::RenderTexture(int Width, int Height, bool wantCubeMap){
	width = Width;
	height = Height;
	channels = 4;
	minFilter = GL_LINEAR_MIPMAP_LINEAR;
	magFilter = GL_LINEAR;
	wrapModeS = GL_CLAMP_TO_EDGE_EXT;
	wrapModeT = GL_CLAMP_TO_EDGE_EXT;
	wrapModeR = GL_CLAMP_TO_EDGE_EXT;
	internalFormat = GL_RGBA8;
	target = (wantCubeMap? GL_TEXTURE_CUBE_MAP_ARB : GL_TEXTURE_2D);
}

RenderTexture::~RenderTexture(){

}

bool RenderTexture::upload(){
	bool created = create();

	glGenTextures(1, &textureID);
	glBindTexture(target, textureID);
	glTexParameteri(target, GL_TEXTURE_MIN_FILTER, minFilter);
	
	glTexParameteri(target, GL_TEXTURE_MAG_FILTER, magFilter);
	
	glTexParameteri(target, GL_TEXTURE_WRAP_S, wrapModeS);
	glTexParameteri(target, GL_TEXTURE_WRAP_T, wrapModeT);
	if (target == GL_TEXTURE_CUBE_MAP_ARB) glTexParameteri(target, GL_TEXTURE_WRAP_R_EXT, wrapModeR);
	
	glTexParameteri(target, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);

/*
	char *tmp = new char[width * height * 4];
	memset(tmp, 255, width * height * 2);
	glTexImage2D(target, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, tmp);
	delete tmp;
*/

	return true;
}

bool RenderTexture::create(){
	hdc = wglGetCurrentDC();
	hglrc = wglGetCurrentContext();

    int nPixelFormat;
	unsigned int nMatchingPixelFormats;

	int iAttributeIList[] = {
		WGL_SUPPORT_OPENGL_ARB,       GL_TRUE, 
		WGL_DRAW_TO_PBUFFER_ARB,      GL_TRUE,
		WGL_BIND_TO_TEXTURE_RGBA_ARB, GL_TRUE,
		WGL_PIXEL_TYPE_ARB,           WGL_TYPE_RGBA_ARB,
		WGL_ACCELERATION_ARB,         WGL_FULL_ACCELERATION_ARB,
		
		WGL_COLOR_BITS_ARB,           32,
		/*WGL_RED_BITS_ARB, 8, // at least 8 bits for RED channel
		WGL_GREEN_BITS_ARB, 8, // at least 8 bits for GREEN channel
		WGL_BLUE_BITS_ARB, 8, // at least 8 bits for BLUE channel
		WGL_ALPHA_BITS_ARB, 8, // at least 8 bits for ALPHA channel
		*/

		WGL_DEPTH_BITS_ARB,           24,
		WGL_STENCIL_BITS_ARB,         0,
		WGL_DOUBLE_BUFFER_ARB,        GL_FALSE,
		0 // End of array
	};

	wglChoosePixelFormatARB(hdc, iAttributeIList, NULL, 1, &nPixelFormat, &nMatchingPixelFormats);

	if (nMatchingPixelFormats < 1){
		::MessageBox(NULL, "Error", "tjo", MB_OK);
	}

	
	int iPBufferAttributes[] = {
		//WGL_PBUFFER_LARGEST_ARB,   GL_FALSE,
		WGL_TEXTURE_FORMAT_ARB,    WGL_TEXTURE_RGBA_ARB,
		WGL_TEXTURE_TARGET_ARB,    (target == GL_TEXTURE_CUBE_MAP_ARB)? WGL_TEXTURE_CUBE_MAP_ARB : WGL_TEXTURE_2D_ARB,
		WGL_MIPMAP_TEXTURE_ARB,    GL_TRUE,
		0
	};

	hPBuffer = wglCreatePbufferARB(hdc, nPixelFormat, width, height, iPBufferAttributes);
	
	hPdc = wglGetPbufferDCARB(hPBuffer);
	hPrc = wglCreateContext(hPdc);

	wglShareLists(wglGetCurrentContext(), hPrc);

	// Other possibility ... check out later
	//hPrc = wglGetCurrentContext();


	return true;
}

bool RenderTexture::unload(){
	if (textureID != TEXTURE_INVALID){
		glDeleteTextures(1,&textureID);
		textureID = 0;
	}

	wglDeleteContext(hPrc);
	wglReleasePbufferDCARB(hPBuffer, hPdc);
	wglDestroyPbufferARB(hPBuffer);
	return true;
}

void RenderTexture::beginRenderToTexture(){
	/*glFinish();
	SetLastError(ERROR_SUCCESS);
	int flag = 0;
	if (wglQueryPbufferARB(hPBuffer, WGL_PBUFFER_LOST_ARB, &flag) == FALSE){
		::MessageBox(NULL, "Trasig pBuffer", "Gone", MB_OK);
	}*/

    wglMakeCurrent(hPdc, hPrc);
    //wglMakeContextCurrentARB(hPdc, hPdc, hPrc);
	/*if (flag != 0) ::MessageBox(NULL, "Lost!", "Gone", MB_OK);
	int err;
	if ((err = GetLastError()) != ERROR_SUCCESS){
		static int t = 0;
		if (t < 10){
			LPVOID lpMsgBuf;
			FormatMessage( 
				FORMAT_MESSAGE_ALLOCATE_BUFFER | 
				FORMAT_MESSAGE_FROM_SYSTEM | 
				FORMAT_MESSAGE_IGNORE_INSERTS,
				NULL,
				err,
				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
				(LPTSTR) &lpMsgBuf,
				0,
				NULL);
			MessageBox(NULL, (LPCTSTR) lpMsgBuf, "Trasig pBuffer", MB_OK | MB_ICONINFORMATION);
			LocalFree(lpMsgBuf);
		}
		t++;
	}*/
}

void RenderTexture::endRenderToTexture(){
	wglMakeCurrent(hdc, hglrc);
    //wglMakeContextCurrentARB(hdc, hdc, hglrc);
}

void RenderTexture::setRenderFace(GLint face){
	int iPBufferAttributes[] = {
		WGL_CUBE_MAP_FACE_ARB,    face,
		0
	};

    //wglSetPbufferAttribARB(hPBuffer, iPBufferAttributes);
}

bool RenderTexture::activate(){
	glBindTexture(target, textureID);

	//glCopyTexSubImage2D(target, 0, 0, 0, 0, 0, width, height);


	wglBindTexImageARB(hPBuffer, WGL_FRONT_LEFT_ARB);
	return true;
}

bool RenderTexture::deactivate(){
	wglReleaseTexImageARB(hPBuffer, WGL_FRONT_LEFT_ARB);
	return true;
}

#endif // _RENDERTEXTURE_H_
