                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/

#ifndef _SPOTLIGHT_H_
#define _SPOTLIGHT_H_

#include "OpenGL/Light.h"

class SpotLight : public Light {
protected:
	float Wx, Wy, Wz;
	float fov;
	float lightView[16];
public:
	void setOrientation(float wx, float wy, float wz){
		Wx = wx;
		Wy = wy;
		Wz = wz;

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glRotatef(Wz, 0.0f, 0.0f, 1.0f);
		glRotatef(Wx, 1.0f, 0.0f, 0.0f);
		glRotatef(Wy, 0.0f, 1.0f, 0.0f);
		glGetFloatv(GL_MODELVIEW_MATRIX, lightView);

		glPopMatrix();
	}

	void setFov(float Fov){
		fov = Fov;
	}

	void getSpace(Vertex &dx, Vertex &dy, Vertex &dz){
		dx = Vertex(lightView[0], lightView[4], lightView[8]);
		dy = Vertex(lightView[1], lightView[5], lightView[9]);
		dz = Vertex(lightView[2], lightView[6], lightView[10]);
	}

	bool isPointInFrontOf(Vertex &v){
		Vertex normal(lightView[2], lightView[6], lightView[10]);

		return (v * normal -normal * position > 0);				
	}

	void setupSpotLightProjection(TextureHandler *texHandler, int tmu, float cutOffDist){
		texHandler->setActiveTexture(tmu);

		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
		glTranslatef(0.5, 0.5, 0.0);
		glScalef(0.5, 0.5, 1.0);
		gluPerspective(fov, 1, 0.01f, 100);
		
		glMultMatrixf(lightView);
		glTranslatef(-position.x, -position.y, -position.z);


		texHandler->setActiveTexture(tmu + 1);
		glMatrixMode(GL_TEXTURE);

		float flipAndScale[16] = {
			0,          0,          cutOffDist, 0,
			0,          cutOffDist, 0,          0,
			cutOffDist, 0,          0,          0,
			0,          0,          0,          1
		};
		glLoadMatrixf(flipAndScale);
		
		glMultMatrixf(lightView);
		glTranslatef(-position.x, -position.y, -position.z);
	}

	void draw(){
		Vertex dx(lightView[0], lightView[4], lightView[8]);
		Vertex dy(lightView[1], lightView[5], lightView[9]);
		Vertex dz(lightView[2], lightView[6], lightView[10]);
		Vertex v;

		float x,y;
		int i;

		glBegin(GL_TRIANGLE_STRIP);
		for (i = 0; i <= 16; i++){
			x = cosf(i * PI / 8.0f);
			y = sinf(i * PI / 8.0f);

			glTexCoord2f(i / 4.0f,0);
			v = position + 20 * (-x * dx + y * dy) - 40 * dz;
			glVertex3fv((GLfloat *) &v);

			glTexCoord2f(i / 4.0f,2);
			v = position + 20 * (-x * dx + y * dy) + 20 * dz;
			glVertex3fv((GLfloat *) &v);
		}
		glEnd();

		glBegin(GL_TRIANGLE_STRIP);
		for (i = 0; i <= 16; i++){
			x = cosf(i * PI / 8.0f);
			y = sinf(i * PI / 8.0f);

			glTexCoord2f(i / 4.0f,0);
			v = position + 20 * (-x * dx + y * dy) + 20 * dz;
			glVertex3fv((GLfloat *) &v);

			glTexCoord4f(i / 6.0f, 2.0f / 3.0f, 0, 2.0f / 3.0f);
			v = position + 30 * (-x * dx + y * dy) + 40 * dz;
			glVertex3fv((GLfloat *) &v);
		}
		glEnd();

		glBegin(GL_TRIANGLE_FAN);
		for (i = 0; i < 16; i++){
			x = cosf(i * PI / 8.0f);
			y = sinf(i * PI / 8.0f);

			glTexCoord2f(0.5f * x,0.5f * y);
			v = position + 20 * (-x * dx + y * dy) - 40 * dz;
			glVertex3fv((GLfloat *) &v);
		}
		glEnd();
	}
};


#endif // _SPOTLIGHT_H_